/* =========================================  \
|  Uses of Row Pattern Matching               |
|  Presentation by Kim Berg Hansen            |
|  Trivadis 2019                              |
|  http://bit.ly/kibeha_patmatch4_pptx        |
|  http://bit.ly/kibeha_patmatch4_sql         |
\  ========================================= */


/* ------------------------  \
|  STOCK TICKER              |
\  ------------------------ */

/*
   Table with stock ticker data
*/

create table ticker (
   symbol varchar2(10)
 , day    date
 , price  number
);
 
insert into ticker values('PLCH', DATE '2011-04-01', 12);
insert into ticker values('PLCH', DATE '2011-04-02', 17);
insert into ticker values('PLCH', DATE '2011-04-03', 19);
insert into ticker values('PLCH', DATE '2011-04-04', 21);
insert into ticker values('PLCH', DATE '2011-04-05', 25);
insert into ticker values('PLCH', DATE '2011-04-06', 12);
insert into ticker values('PLCH', DATE '2011-04-07', 15);
insert into ticker values('PLCH', DATE '2011-04-08', 20);
insert into ticker values('PLCH', DATE '2011-04-09', 24);
insert into ticker values('PLCH', DATE '2011-04-10', 25);
insert into ticker values('PLCH', DATE '2011-04-11', 19);
insert into ticker values('PLCH', DATE '2011-04-12', 15);
insert into ticker values('PLCH', DATE '2011-04-13', 25);
insert into ticker values('PLCH', DATE '2011-04-14', 25);
insert into ticker values('PLCH', DATE '2011-04-15', 14);
insert into ticker values('PLCH', DATE '2011-04-16', 12);
insert into ticker values('PLCH', DATE '2011-04-17', 14);
insert into ticker values('PLCH', DATE '2011-04-18', 24);
insert into ticker values('PLCH', DATE '2011-04-19', 23);
insert into ticker values('PLCH', DATE '2011-04-20', 22);
commit;

/*
   Example from data warehousing guide
*/

select *
from ticker match_recognize (
     partition by symbol
     order by day
     measures  strt.day as start_day,
               final last(down.day) as bottom_day,
               final last(up.day) as end_day,
               match_number() as match_num,
               classifier() as var_match
     all rows per match
     after match skip to last up
     pattern (strt down+ up+)
     define
        down as down.price < prev(down.price),
        up as up.price > prev(up.price)
     ) mr
order by mr.symbol, mr.match_num, mr.day;

/*
   Similar with ONE ROW PER MATCH
*/

select *
from ticker match_recognize (
     partition by symbol
     order by day
     measures  strt.day as start_day,
               final last(down.day) as bottom_day,
               final last(down.price) as bottom_price,
               final last(up.day) as end_day,
               match_number() as match_num
     one row per match
     after match skip to last up
     pattern (strt down+ up+)
     define
        down as down.price < prev(down.price),
        up as up.price > prev(up.price)
     ) mr
order by mr.symbol, mr.match_num;

/*
   Example from plsqlchallenge.oracle.com championship quiz
*/

select symbol, day, price, up_day, up_avg, up_total
  from ticker
match_recognize (
   partition by symbol
   order by day
   measures
      final count(up.*) as days_up
    , up.price - prev(up.price) as up_day
    , (final last(up.price) - strt.price) / final count(up.*)
         as up_avg
    , up.price - strt.price as up_total
   all rows per match
   after match skip to last up
   pattern (
      strt up+
   )
   define
      up as up.price > prev(up.price)
)
 order by day;


/*
   Cleanup
*/

drop table ticker purge;


/* ------------------------  \
|  GROUPING SEQUENCES        |
\  ------------------------ */

-- Example by Stew Ashton
-- Blogged in: https://stewashton.wordpress.com/2014/03/05/12c-match_recognize-grouping-sequences/

-- Table of numeric values in some sequential groups.

create table ex1 (numval)
as
select  1 from dual union all
select  2 from dual union all
select  3 from dual union all
select  5 from dual union all
select  6 from dual union all
select  7 from dual union all
select 10 from dual union all
select 11 from dual union all
select 12 from dual union all
select 20 from dual;

-- Define states "b" row must have numval exactly 1 greater than previous row.
-- Pattern matches any row to "a" row and then zero or more "b" rows,
-- so match continues until a row that no longer is exactly 1 greater.

select *
  from ex1
match_recognize (
   order by numval
   measures
      first(numval) firstval
    , last(numval) lastval
    , count(*) cnt
   pattern (
      a b*
   )
   define
      b as numval = prev(numval) + 1
);

-- Tabibitosan method by Aketi Jyuuzou

select min(numval) firstval
     , max(numval) lastval
     , count(*) cnt
  from (
   select numval
        , numval-row_number() over (order by numval) as grp
   from ex1
  )
 group by grp
 order by min(numval);

-- Cleanup
drop table ex1 purge;


/* ------------------------  \
|  MERGE DATE RANGES         |
\  ------------------------ */

-- Example originally by Stew Ashton - adapted slightly with other data
-- Blogged in: https://stewashton.wordpress.com/2015/06/10/merging-overlapping-date-ranges-with-match_recognize/

-- Table of date ranges

create table t ( id int, start_date date, end_date date );

insert into t values ( 1, date '2014-01-01', date '2014-01-03');
insert into t values ( 2, date '2014-01-02', date '2014-01-05');
insert into t values ( 3, date '2014-01-02', date '2014-01-06');
insert into t values ( 4, date '2014-01-03', date '2014-01-05');
insert into t values ( 5, date '2014-01-05', date '2014-01-07');
insert into t values ( 6, date '2014-01-23', date '2014-02-01');
insert into t values ( 7, date '2014-01-25', date '2014-02-01');
insert into t values ( 8, date '2014-02-01', date '2014-02-10');
insert into t values ( 9, date '2014-02-01', date '2014-02-04');
insert into t values (10, date '2014-02-05', date '2014-02-12');
insert into t values (11, date '2014-02-10', date '2014-02-15');
commit;

-- Two date ranges are contiguous, if start of second equal end of first.
-- Task is to merge contiguous or overlapping date ranges.

select *
  from t
match_recognize(
   order by start_date, end_date
   measures
      first(start_date) start_date
    , max(end_date) end_date
    , count(*) c
   pattern(
      a* b
   )
   define
      a as next(start_date) <= max(end_date)
);

-- Handling when NULL values mean infinity.

insert into t values (12, null, date '2014-01-01');
insert into t values (13, null, date '2014-01-02');
insert into t values (14, date '2014-02-19', date '2014-02-21');
insert into t values (14, date '2014-02-20', null);
insert into t values (15, date '2014-02-21', null);
commit;

select *
  from t
match_recognize(
   order by start_date nulls first, end_date nulls last
   measures
      first(start_date) start_date
    , nullif(max(nvl(end_date, date '9999-12-31')), date '9999-12-31') end_date
    , count(*) c
   pattern(
      a* b
   )
   define
      a as nvl(next(start_date), date '-4712-01-01') <= max(nvl(end_date, date '9999-12-31'))
);

-- Cleanup
drop table t purge;

/* ------------------------  \
|  TABLESPACE GROWTH         |
\  ------------------------ */

-- Quiz by Kim Berg Hansen on plsqlchallenge.oracle.com

-- Table with daily midnight logging of tablespace used size

create table plch_space (
   tabspace    varchar2(30)
 , sampledate  date
 , gigabytes   number
);

insert into plch_space values ('MYSPACE'  , date '2014-02-01', 100);
insert into plch_space values ('MYSPACE'  , date '2014-02-02', 103);
insert into plch_space values ('MYSPACE'  , date '2014-02-03', 116);
insert into plch_space values ('MYSPACE'  , date '2014-02-04', 129);
insert into plch_space values ('MYSPACE'  , date '2014-02-05', 142);
insert into plch_space values ('MYSPACE'  , date '2014-02-06', 160);
insert into plch_space values ('MYSPACE'  , date '2014-02-07', 165);
insert into plch_space values ('MYSPACE'  , date '2014-02-08', 210);
insert into plch_space values ('MYSPACE'  , date '2014-02-09', 230);
insert into plch_space values ('MYSPACE'  , date '2014-02-10', 239);
insert into plch_space values ('YOURSPACE', date '2014-02-06',  50);
insert into plch_space values ('YOURSPACE', date '2014-02-07',  53);
insert into plch_space values ('YOURSPACE', date '2014-02-08',  72);
insert into plch_space values ('YOURSPACE', date '2014-02-09',  97);
insert into plch_space values ('YOURSPACE', date '2014-02-10', 101);
insert into plch_space values ('HISSPACE',  date '2014-02-06', 100);
insert into plch_space values ('HISSPACE',  date '2014-02-07', 130);
insert into plch_space values ('HISSPACE',  date '2014-02-08', 145);
insert into plch_space values ('HISSPACE',  date '2014-02-09', 200);
insert into plch_space values ('HISSPACE',  date '2014-02-10', 225);
insert into plch_space values ('HISSPACE',  date '2014-02-11', 255);
insert into plch_space values ('HISSPACE',  date '2014-02-12', 285);
insert into plch_space values ('HISSPACE',  date '2014-02-13', 315);
commit;

-- If a tablespace grows >= 25% in one day, that's a FAST growth spurt.
-- If a tablespace grows >= 10% and <25% in one day, that's a SLOW growth spurt.
-- We want to see all FAST growth spurts as that might indicate problems.
-- But also periods with at least 3 consecutive SLOW growth spurts.
-- And show average gigabytes grown per day within the periods.


-- Comparing a rows gigabytes with next rows gigabytes give ratio how much growth in that day.
-- Define "fast" and "slow" rows by ratio interval.
-- A "fast" growth spurt is at least 1 "fast" row => fast+
-- A "slow" growth spurt is at least 3 "slow" rows => slow{3,}
-- Either "fast" or "slow" growth spurt using | meaning "or" => fast+ | slow{3,}
-- classifier() function tells whether row was matched "fast" or "slow".
-- Note: in "one row per match" classifier() gives classifier for last row of match.

select tabspace
     , spurttype
     , startdate
     , startgb
     , enddate
     , endgb
     , (endgb - startgb) / days avg_daily_gb
  from plch_space
match_recognize (
   partition by tabspace
   order by sampledate
   measures
      classifier() as spurttype
    , first(sampledate) as startdate
    , first(gigabytes) as startgb
    , last(sampledate) as enddate
    , next(gigabytes) as endgb
    , count(*) as days
   one row per match
   after match skip past last row
   pattern (
      fast+ | slow{3,}
   )
   define
      fast as next(fast.gigabytes) / fast.gigabytes >= 1.25
    , slow as next(slow.gigabytes) / slow.gigabytes >= 1.10 and
              next(slow.gigabytes) / slow.gigabytes <  1.25
)
 order by tabspace
        , startdate;

-- Same definitions, but puts avg_daily_gb calculation within measures clause.

select tabspace
     , spurttype
     , startdate
     , startgb
     , enddate
     , endgb
     , avg_daily_gb
  from plch_space
match_recognize (
   partition by tabspace
   order by sampledate
   measures
      classifier() as spurttype
    , first(sampledate) as startdate
    , first(gigabytes) as startgb
    , last(sampledate) as enddate
    , next(gigabytes) as endgb
    , (next(gigabytes) - first(gigabytes)) / count(*) as avg_daily_gb
   one row per match
   after match skip past last row
   pattern (
      fast+ | slow{3,}
   )
   define
      fast as next(gigabytes) / gigabytes >= 1.25
    , slow as next(slow.gigabytes) / slow.gigabytes >= 1.10 and
              next(slow.gigabytes) / slow.gigabytes <  1.25
)
 order by tabspace
        , startdate;

-- Alternative without MATCH_RECOGNIZE that will work in pre-12c databases.

select tabspace
     , spurttype
     , startdate
     , min(gigabytes) keep (
          dense_rank first
          order by sampledate
       ) startgb
     , max(sampledate) enddate
     , max(nextgb) keep (
          dense_rank last
          order by sampledate
       ) endgb
     , avg(daily_gb) avg_daily_gb
  from (
   select tabspace
        , spurttype
        , sampledate
        , gigabytes
        , nextgb
        , daily_gb
        , last_value(spurtstartdate ignore nulls) over (
             partition by tabspace, spurttype
             order by sampledate
             rows between unbounded preceding and current row
          ) startdate
     from (
      select tabspace
           , spurttype
           , sampledate
           , gigabytes
           , nextgb
           , daily_gb
           , case
                when spurttype is not null and
                     ( lag(spurttype) over (
                          partition by tabspace
                          order by sampledate
                       ) is null
                       or
                       lag(spurttype) over (
                          partition by tabspace
                          order by sampledate
                       ) != spurttype
                     )
                then sampledate
             end spurtstartdate
        from (
         select tabspace
              , sampledate
              , gigabytes
              , nextgb
              , nextgb - gigabytes daily_gb
              , case
                   when nextgb >= gigabytes * 1.25 then 'FAST'
                   when nextgb >= gigabytes * 1.10 then 'SLOW'
                end spurttype
           from (
            select tabspace
                 , sampledate
                 , gigabytes
                 , lead(gigabytes) over (
                      partition by tabspace
                      order by sampledate
                   ) nextgb
              from plch_space
                )
             )
          )
    where spurttype is not null
       )
 group by tabspace
        , spurttype
        , startdate
having count(*) >= case spurttype
                      when 'FAST' then 1
                      when 'SLOW' then 3
                   end 
 order by tabspace
        , startdate;

-- Cleanup
drop table plch_space purge;


/* ------------------------  \
|  BINS WITH LIMIT CAPACITY  |
\  ------------------------ */

-- Example by Stew Ashton
-- Blogged in: https://stewashton.wordpress.com/2014/03/03/database-12c-match_recognize-for-all-sizes-of-data/

create table t (study_site number, cnt number);

insert into t (study_site,cnt) values (1001,3407);
insert into t (study_site,cnt) values (1002,4323);
insert into t (study_site,cnt) values (1004,1623);
insert into t (study_site,cnt) values (1008,1991);
insert into t (study_site,cnt) values (1011,885);
insert into t (study_site,cnt) values (1012,11597);
insert into t (study_site,cnt) values (1014,1989);
insert into t (study_site,cnt) values (1015,5282);
insert into t (study_site,cnt) values (1017,2841);
insert into t (study_site,cnt) values (1018,5183);
insert into t (study_site,cnt) values (1020,6176);
insert into t (study_site,cnt) values (1022,2784);
insert into t (study_site,cnt) values (1023,25865);
insert into t (study_site,cnt) values (1024,3734);
insert into t (study_site,cnt) values (1026,137);
insert into t (study_site,cnt) values (1028,6005);
insert into t (study_site,cnt) values (1029,76);
insert into t (study_site,cnt) values (1031,4599);
insert into t (study_site,cnt) values (1032,1989);
insert into t (study_site,cnt) values (1034,3427);
insert into t (study_site,cnt) values (1036,879);
insert into t (study_site,cnt) values (1038,6485);
insert into t (study_site,cnt) values (1039,3);
insert into t (study_site,cnt) values (1040,1105);
insert into t (study_site,cnt) values (1041,6460);
insert into t (study_site,cnt) values (1042,968);
insert into t (study_site,cnt) values (1044,471);
insert into t (study_site,cnt) values (1045,3360);
commit;

-- Create groups of consecutive study_site with sum(cnt) at most 65.000
-- Aggregate SUM in Define is "running", so pattern of "a+" continues matching
-- as long as rolling sum(cnt) is <= 65.000.

select *
  from t
match_recognize (
   order by study_site
   measures
      first(study_site) first_site
    , last(study_site) last_site
    , sum(cnt) sum_cnt
   one row per match
   after match skip past last row
   pattern (
      a+
   )
   define
      a as sum(cnt) <= 65000
);

-- Order by CNT descending

select *
  from t
match_recognize (
   order by cnt desc, study_site
   measures
      count(*) sites
    , sum(cnt) sum_cnt
    , min(cnt) min_cnt
    , max(cnt) max_cnt
   one row per match
   after match skip past last row
   pattern (
      a+
   )
   define
      a as sum(cnt) <= 65000
);

-- Order large/small CNT interleaved

select *
  from (
   select study_site, cnt
        , least(
             row_number() over (
                order by cnt
             )
           , row_number() over (
                order by cnt desc
             )
          ) rn
     from t
  )
match_recognize (
   order by rn, cnt desc, study_site
   measures
      count(*) sites
    , sum(cnt) sum_cnt
    , min(cnt) min_cnt
    , max(cnt) max_cnt
   one row per match
   after match skip past last row
   pattern (
      a+
   )
   define
      a as sum(cnt) <= 65000
);

-- Cleanup
drop table t purge;


/* ------------------------  \
|  LIMIT NUMBER OF BINS      |
\  ------------------------ */

-- Example by Stew Ashton
-- Blogged in: https://stewashton.wordpress.com/2014/06/06/bin-fitting-problems-with-sql/

create table items
as
select level item_name, level item_value 
  from dual
connect by level <= 10;

select *
  from items
 order by item_name;
 
-- Bin fitting to fixed number of bins - here 3 bins.
-- Try to fill the bins as equally as possible.

-- Algorithm:
-- - First, order the items by value in descending order.
-- - Then, assign each item to whatever bin has the smallest sum so far.

-- Pattern matches either bin1 or bin2 or bin3 - zero or more occurences.
-- A row will be bin1 if it is the first (only) row in bin1 or the sum of values
-- in bin1 previously (the running sum includes current row so we subtract it)
-- is less than or equal to the smallest sum in either bin2 or bin3.
-- If a row is not bin1, it will be bin2 if it is the first (only) row in bin2 or
-- the sum of values in bin2 previously is <= the sum in bin3.
-- If a row is not bin2, it will always be bin3.
-- So the match here will always continue to the end. If we used "one row per match"
-- we would get a single row output. The running sums here show how the bins fill up.

select *
  from items
match_recognize (
   order by item_value desc
   measures
      to_number(substr(classifier(),4)) bin#,
      sum(bin1.item_value) bin1,
      sum(bin2.item_value) bin2,
      sum(bin3.item_value) bin3
   all rows per match
   pattern (
      (bin1|bin2|bin3)*
   )
   define
      bin1 as count(bin1.*) = 1
           or sum(bin1.item_value)-bin1.item_value
                <= least(sum(bin2.item_value), sum(bin3.item_value))
    , bin2 as count(bin2.*) = 1
           or sum(bin2.item_value)-bin2.item_value
                <= sum(bin3.item_value)
);

-- Cleanup
drop table items purge;


/* ------------------------  \
|  HIERARCHY CHILD COUNT     |
\  ------------------------ */

-- Blog by Kim Berg Hansen
-- http://www.kibeha.dk/2015/07/row-pattern-matching-nested-within.html

-- Uses standard SCOTT.EMP table.

-- Regular scalar subquery method.

select empno
     , lpad(' ', (level-1)*2) || ename as ename
     , (
         select count(*)
           from emp sub
          start with sub.mgr = emp.empno
          connect by sub.mgr = prior sub.empno
       ) subs
  from emp
 start with mgr is null
 connect by mgr = prior empno
 order siblings by empno;

-- Same result with pattern matching.
-- WITH clause for clarity, inline view in WITH to preserve ordering with ROWNUM.
-- Define "higher" as row level is higher than starting row => is child/grandchild/.. of starting row.
-- Any row matches "strt", then match zero or more "higher" rows.
-- Data from "strt" row is the employee we output in "one row per match".
-- "higher" data will be null on "strt" rows, so count(higher.lvl) does not include "strt" row.
-- "after match ship to next row" tries to find a new match at next row after "strt" row.

with hierarchy as (
   select lvl, empno, ename, rownum as rn
   from (
      select level as lvl, empno, ename
        from emp
       start with mgr is null
       connect by mgr = prior empno
       order siblings by empno
   )
)
select empno
     , lpad(' ', (lvl-1)*2) || ename as ename
     , subs
  from hierarchy
match_recognize (
   order by rn
   measures
      strt.rn as rn
    , strt.lvl as lvl
    , strt.empno as empno
    , strt.ename as ename
    , count(higher.lvl) as subs
   one row per match
   after match skip to next row
   pattern (
      strt higher*
   )
   define
      higher as higher.lvl > strt.lvl
)
 order by rn;

-- To demonstrate what happens, here is output "all rows per match".
-- match_number() function is a consecutive numbering of the matches found.
-- classisifer() function shows whether it is a "strt" row or "higher" row.
-- When using "all rows per match", count(higher.lvl) becomes a rolling count.
-- Need final keyword to get total count even on all rows.
-- Without final (=running), similar to rows between unbounded preceding and current row.
-- With final, similar to rows between unbounded preceding and unbounded following.

with hierarchy as (
   select lvl, empno, ename, rownum as rn
   from (
      select level as lvl, empno, ename
        from emp
       start with mgr is null
       connect by mgr = prior empno
       order siblings by empno
   )
)
select mn
     , rn
     , empno
     , lpad(' ', (lvl-1)*2) || ename as ename
     , roll
     , subs
     , cls
     , stno
     , stname
     , hino
     , hiname
  from hierarchy
match_recognize (
   order by rn
   measures
      match_number() as mn
    , classifier() as cls
    , strt.empno as stno
    , strt.ename as stname
    , higher.empno as hino
    , higher.ename as hiname
    , count(higher.lvl) as roll
    , final count(higher.lvl) as subs
   all rows per match
   after match skip to next row
   pattern (
      strt higher*
   )
   define
      higher as higher.lvl > strt.lvl
)
 order by mn, rn;

-- Pivoting visualizes which rows each match_number() matches.

with hierarchy as (
   select lvl, empno, ename, rownum as rn
   from (
      select level as lvl, empno, ename
        from emp
       start with mgr is null
       connect by mgr = prior empno
       order siblings by empno
   )
)
select rn, empno, ename
     , case "1"  when 1 then 'XX' end "1" 
     , case "2"  when 1 then 'XX' end "2" 
     , case "3"  when 1 then 'XX' end "3" 
     , case "4"  when 1 then 'XX' end "4" 
     , case "5"  when 1 then 'XX' end "5" 
     , case "6"  when 1 then 'XX' end "6" 
     , case "7"  when 1 then 'XX' end "7" 
     , case "8"  when 1 then 'XX' end "8" 
     , case "9"  when 1 then 'XX' end "9" 
     , case "10" when 1 then 'XX' end "10"
     , case "11" when 1 then 'XX' end "11"
     , case "12" when 1 then 'XX' end "12"
     , case "13" when 1 then 'XX' end "13"
     , case "14" when 1 then 'XX' end "14"
  from (
   select mn
        , rn
        , empno
        , lpad(' ', (lvl-1)*2) || ename as ename
     from hierarchy
   match_recognize (
      order by rn
      measures
         match_number() as mn
      all rows per match
      after match skip to next row
      pattern (
         strt higher*
      )
      define
         higher as higher.lvl > strt.lvl
   )
  )
 pivot (
   count(*)
   for mn in (
      1,2,3,4,5,6,7,8,9,10,11,12,13,14
   )
 )
 order by rn;

-- When using "one row per match" can still include some more data
-- using first, last or aggregate functions, like here max(higher.lvl)

with hierarchy as (
   select lvl, empno, ename, rownum as rn
   from (
      select level as lvl, empno, ename
        from emp
       start with mgr is null
       connect by mgr = prior empno
       order siblings by empno
   )
)
select empno
     , lpad(' ', (lvl-1)*2) || ename as ename
     , subs
     , hifrom
     , hito
     , himax
  from hierarchy
match_recognize (
   order by rn
   measures
      strt.rn as rn
    , strt.lvl as lvl
    , strt.empno as empno
    , strt.ename as ename
    , count(higher.lvl) as subs
    , first(higher.ename) as hifrom
    , last(higher.ename) as hito
    , max(higher.lvl) as himax
   one row per match
   after match skip to next row
   pattern (
      strt higher*
   )
   define
      higher as higher.lvl > strt.lvl
)
 order by rn;

-- Only want to output those with subordinates?
-- Could put in inline view and filter on "subs > 0".
-- Better just to change "strt higher*" to "strt higher+".
-- That requires at least one "higher" row (=subordinate) to get a match.

with hierarchy as (
   select lvl, empno, ename, rownum as rn
   from (
      select level as lvl, empno, ename
        from emp
       start with mgr is null
       connect by mgr = prior empno
       order siblings by empno
   )
)
select empno
     , lpad(' ', (lvl-1)*2) || ename as ename
     , subs
  from hierarchy
match_recognize (
   order by rn
   measures
      strt.lvl as lvl
    , strt.empno as empno
    , strt.ename as ename
    , count(higher.lvl) as subs
   one row per match
   after match skip to next row
   pattern (
      strt higher+
   )
   define
      higher as higher.lvl > strt.lvl
);

-- Test on a bigger table

create table bigemp
as
select 1 empno
     , 'LARRY' ename
     , cast(null as number) mgr
  from dual
union all
select dum.dum*10000+empno empno
     , ename || '#' || dum.dum ename
     , coalesce(
          dum.dum*10000+mgr
        , 1
       ) mgr
  from emp
 cross join (
   select level dum
     from dual
   connect by level <= 1000
 ) dum;

create index bigemp_mgr_eno on bigemp (mgr, empno);

begin
   dbms_stats.gather_table_stats(USER,'BIGEMP',cascade=>true);
end;
/

set autotrace traceonly
set timing on

select empno
     , lpad(' ', (level-1)*2) || ename as ename
     , (
         select count(*)
           from bigemp sub
          start with sub.mgr = bigemp.empno
          connect by sub.mgr = prior sub.empno
       ) subs
  from bigemp
 start with mgr is null
 connect by mgr = prior empno
 order siblings by empno;

with hierarchy as (
   select lvl, empno, ename, rownum as rn
   from (
      select level as lvl, empno, ename
        from bigemp
       start with mgr is null
       connect by mgr = prior empno
       order siblings by empno
   )
)
select empno
     , lpad(' ', (lvl-1)*2) || ename as ename
     , subs
  from hierarchy
match_recognize (
   order by rn
   measures
      strt.lvl as lvl
    , strt.empno as empno
    , strt.ename as ename
    , count(higher.lvl) as subs
   one row per match
   after match skip to next row
   pattern (
      strt higher*
   )
   define
      higher as higher.lvl > strt.lvl
);

-- Cleanup
drop table bigemp purge;


/* =========================================  \
|  End of script                              |
\  ========================================= */