/* =========================================  \
|  Data Twisting                              |
|  Presentation by Kim Berg Hansen            |
|  Trivadis 2019                              |
|  http://bit.ly/kibeha_datatwist4_pptx       |
|  http://bit.ly/kibeha_datatwist4_sql        |
|  TWIST, SHAKE, RATTLE 'n' ROLL              |
\  ========================================= */

/* ------------------------  \
|  TWIST - Single dimension  |
\  ------------------------ */

/*
   Table with test data
*/
create table sales1 (
   category varchar2(10)
 , emea     number
 , amer     number
 , asoc     number
);

insert into sales1 values ('Beer', 200000, 150000, 225000);
insert into sales1 values ('Wine',  10000,  25000,  17500);

commit;

select *
  from sales1
 order by category;

/*
   UNPIVOT
*/
select category, region, sales
  from sales1
unpivot (
   sales
   for region in (
      emea as 'EMEA'
    , amer as 'AMER'
    , asoc as 'ASOC'
   )
)
 order by category, region;

/*
   With row generator numeric
*/
select category
     , case n# when 1 then 'EMEA' when 2 then 'AMER' when 3 then 'ASOC' end region
     , case n# when 1 then  emea  when 2 then  amer  when 3 then  asoc  end sales
  from sales1
 cross join (
   select level n# from dual connect by level <= 3
 )
 order by category, region;

/*
   With row generator dimension style
*/
with r (region) as (
   select 'EMEA' from dual union all
   select 'AMER' from dual union all
   select 'ASOC' from dual
)
select category
     , region
     , case region
          when 'EMEA' then emea
          when 'AMER' then amer
          when 'ASOC' then asoc
       end sales
  from sales1
 cross join r
 order by category, region;

/*
   Cleanup
*/
drop table sales1 purge;


/* ------------------------  \
|  TWIST - Multi dimension   |
\  ------------------------ */

/*
   Table with test data
*/
create table sales2 (
   category       varchar2(10)
 , dk_b2b_qty     number
 , dk_b2b_amount  number
 , dk_b2c_qty     number
 , dk_b2c_amount  number
 , uk_b2b_qty     number
 , uk_b2b_amount  number
 , uk_b2c_qty     number
 , uk_b2c_amount  number
);

insert into sales2 values ('Beer', 500, 5000, 250, 2500, 100, 1000, 200, 2000);
insert into sales2 values ('Wine', 150, 3000, 200, 4000, 400, 8000, 300, 6000);

commit;

select *
  from sales2
 order by category;

/*
   UNPIVOT two measures, two dimensions
*/
select category, country, channel, qty, amount
  from sales2
unpivot (
   ( qty, amount )
   for ( country, channel )
   in (
      (dk_b2b_qty, dk_b2b_amount) as ('DK', 'B2B')
    , (dk_b2c_qty, dk_b2c_amount) as ('DK', 'B2C')
    , (uk_b2b_qty, uk_b2b_amount) as ('UK', 'B2B')
    , (uk_b2c_qty, uk_b2c_amount) as ('UK', 'B2C')
   )
)
 order by category, country, channel;

/*
   UNPIVOT two measures, one dimension
*/
select category, country_and_channel, qty, amount
  from sales2
unpivot (
   ( qty, amount )
   for ( country_and_channel )
   in (
      (dk_b2b_qty, dk_b2b_amount) as ('DK_B2B')
    , (dk_b2c_qty, dk_b2c_amount) as ('DK_B2C')
    , (uk_b2b_qty, uk_b2b_amount) as ('UK_B2B')
    , (uk_b2c_qty, uk_b2c_amount) as ('UK_B2C')
   )
)
 order by category, country_and_channel;

/*
   UNPIVOT one measure, two dimensions
*/
select category, country, channel, amount
  from sales2
unpivot (
   ( amount )
   for ( country, channel )
   in (
      (dk_b2b_amount) as ('DK', 'B2B')
    , (dk_b2c_amount) as ('DK', 'B2C')
    , (uk_b2b_amount) as ('UK', 'B2B')
    , (uk_b2c_amount) as ('UK', 'B2C')
   )
)
 order by category, country, channel;

/*
   Cleanup
*/
drop table sales2 purge;


/* ------------------------  \
|  SHAKE - Single dimension  |
\  ------------------------ */

/*
   Table with test data
*/
create table sales3 (
   category varchar2(10)
 , region   varchar2(10)
 , sales    number
);

insert into sales3 values ('Beer', 'EMEA', 200000);
insert into sales3 values ('Beer', 'AMER', 150000);
insert into sales3 values ('Beer', 'ASOC', 225000);
insert into sales3 values ('Wine', 'EMEA',  10000);
insert into sales3 values ('Wine', 'AMER',  25000);
insert into sales3 values ('Wine', 'ASOC',  17500);

commit;

select *
  from sales3
 order by category, region;

/*
   PIVOT
*/
select category, emea, amer, asoc
  from sales3
 pivot (
   sum(sales)
   for region in (
      'EMEA' as emea
    , 'AMER' as amer
    , 'ASOC' as asoc
   )
 )
 order by category;

/*
   With GROUP BY
*/
select category
     , sum(case region
              when 'EMEA' then sales
           end) as emea
     , sum(case region
              when 'AMER' then sales
           end) as amer
     , sum(case region
              when 'ASOC' then sales
           end) as asoc
  from sales3
 group by category
 order by category;

/*
   Aggregation when non-unique
*/
insert into sales3 values ('Beer', 'AMER',  25000);

commit;

select *
  from sales3
 order by category, region;

/*
   Same PIVOT as before
*/
select category, emea, amer, asoc
  from sales3
 pivot (
   sum(sales)
   for region in (
      'EMEA' as emea
    , 'AMER' as amer
    , 'ASOC' as asoc
   )
 )
 order by category;

/*
   Add a COUNT for multiple measures
   This raises ORA-00918: column ambiguously defined
*/
select *
  from sales3
 pivot (
   sum(sales)
 , count(*)
   for region in (
      'EMEA' as emea
    , 'AMER' as amer
    , 'ASOC' as asoc
   )
 )
 order by category;

/*
   Add measure aliases
*/
select category
     , emea_sale, emea_cnt
     , amer_sale, amer_cnt
     , asoc_sale, asoc_cnt
  from sales3
 pivot (
   sum(sales) as sale
 , count(*)   as cnt
   for region in (
      'EMEA' as emea
    , 'AMER' as amer
    , 'ASOC' as asoc
   )
 )
 order by category;

/*
   Exactly one measure may be left without alias
*/
select category
     , emea, emea_cnt
     , amer, amer_cnt
     , asoc, asoc_cnt
  from sales3
 pivot (
   sum(sales)
 , count(*)   as cnt
   for region in (
      'EMEA' as emea
    , 'AMER' as amer
    , 'ASOC' as asoc
   )
 )
 order by category;

/*
   Cleanup
*/
drop table sales3 purge;


/* ------------------------  \
|  SHAKE - Multi dimension   |
\  ------------------------ */

/*
   Table with test data
*/
create table sales4 (
   category varchar2(10)
 , country  varchar2(10)
 , channel  varchar2(10)
 , qty      number
 , amount   number
);

insert into sales4 values('Beer', 'DK', 'B2B', 500, 5000);
insert into sales4 values('Beer', 'DK', 'B2C', 250, 2500);
insert into sales4 values('Beer', 'UK', 'B2B', 100, 1000);
insert into sales4 values('Beer', 'UK', 'B2C', 200, 2000);
insert into sales4 values('Wine', 'DK', 'B2B', 150, 3000);
insert into sales4 values('Wine', 'DK', 'B2C', 200, 4000);
insert into sales4 values('Wine', 'UK', 'B2B', 400, 8000);
insert into sales4 values('Wine', 'UK', 'B2C', 300, 6000);

commit;

select *
  from sales4
 order by category, country, channel;

/*
   PIVOT
*/
select category
     , dk_b2b_qty, dk_b2b_amount
     , dk_b2c_qty, dk_b2c_amount
     , uk_b2b_qty, uk_b2b_amount
     , uk_b2c_qty, uk_b2c_amount
  from sales4
 pivot (
   sum(qty)    as qty
 , sum(amount) as amount
   for ( country, channel ) in (
      ('DK', 'B2B') as dk_b2b
    , ('DK', 'B2C') as dk_b2c
    , ('UK', 'B2B') as uk_b2b
    , ('UK', 'B2C') as uk_b2c
   )
 )
 order by category;

/*
   Cleanup
*/
drop table sales4 purge;


/* ------------------------  \
|  RATTLE to columns         |
\  ------------------------ */

/*
   Table with test data
*/
create table sales5 (
   txt      varchar2(100)
);

insert into sales5 values ('Beer;200000;150000;225000');
insert into sales5 values ('Wine;10000;25000;17500');

commit;

select *
  from sales5
 order by txt;

/*
   SUBSTR and INSTR
*/
select substr(txt, 1, instr(txt, ';') - 1) category
     , substr(
          txt
        , instr(txt, ';') + 1
        , instr(txt, ';', 1, 2) - instr(txt, ';') -1
       ) emea
     , substr(
          txt
        , instr(txt, ';', 1, 2) + 1
        , instr(txt, ';' , 1, 3) - instr(txt, ';', 1, 2) - 1
       ) amer
     , substr(txt, instr(txt, ';', 1, 3) + 1) asoc
  from sales5
 order by category;

/*
   REGEXP_SUBSTR
*/
select regexp_substr(txt, '[^;]+', 1, 1) category
     , regexp_substr(txt, '[^;]+', 1, 2) emea
     , regexp_substr(txt, '[^;]+', 1, 3) amer
     , regexp_substr(txt, '[^;]+', 1, 4) asoc
  from sales5
 order by category;

/*
   Cleanup
*/
drop table sales5 purge;


/* ------------------------  \
|  RATTLE to rows            |
\  ------------------------ */

/*
   Table with test data
*/
create table beverages1 (
   category varchar2(10)
 , typelist varchar2(100)
);

insert into beverages1 values ('Beer', 'Pilsner;Ale;Stout');
insert into beverages1 values ('Wine', 'Red;Champagne');

commit;

select *
  from beverages1
 order by category;

/*
   Collection from pipelined table function
*/
create type beverage_collection_type as table of varchar2(10);
/

create or replace function beverage_typelist_to_coll (
   typelist  in  beverages1.typelist%type
)
   return beverage_collection_type pipelined
is
   list_len pls_integer;
   from_pos pls_integer;
   to_pos   pls_integer;
begin
   list_len := length(typelist);
   from_pos := 1;
   loop
      to_pos := nvl(nullif(instr(typelist, ';', from_pos), 0), list_len+1);
      pipe row (substr(typelist, from_pos, to_pos-from_pos));
      exit when to_pos > list_len;
      from_pos := to_pos + 1;
   end loop;
end beverage_typelist_to_coll;
/

select category
     , column_value as beverage_type
  from beverages1
     , table(beverage_typelist_to_coll(typelist))
 order by category, beverage_type;

/*
   Using procedure from APEX_UTIL package
*/

declare
   beverage_arr apex_application_global.vc_arr2;
begin
   for b in (
      select category, typelist
        from beverages1
       order by category
   ) loop
      beverage_arr := apex_util.string_to_table(b.typelist, ';');
      for i in 1..beverage_arr.count loop
         dbms_output.put_line(b.category||' - '||beverage_arr(i));
      end loop;
   end loop;
end;
/

/*
   Using function from APEX_STRING package
*/

select category
     , column_value as beverage_type
  from beverages1
     , table(apex_string.split(typelist, ';'))
 order by category, beverage_type;

/*
   Generate rows as many as delimiter count
*/

select category
     , regexp_substr(typelist, '[^;]+', 1, sub#) beverage_type
  from beverages1
 cross join lateral (
   select level sub#
     from dual
   connect by level <= regexp_count(typelist, ';') + 1
 )
 order by category, beverage_type;

/*
   Cleanup
*/
drop function beverage_typelist_to_coll;
drop type beverage_collection_type;
drop table beverages1 purge;


/* ------------------------  \
|  RATTLE to rows with ODCI  |
\  ------------------------ */

/*
   Table with test data
*/
create table beverages2 (
   category varchar2(10)
 , typelist varchar2(100)
);

insert into beverages2 values ('Beer', 'Pilsner|Light;Ale|Medium;Stout|Dark');
insert into beverages2 values ('Wine', 'Red|Red;Champagne|Clear');

commit;

select *
  from beverages2
 order by category;

/*
   ODCI implementation
*/

create or replace type delimited_col_row as object (
 
   g_buffer          varchar2(32000)
 , g_type            anytype
 , g_numcols         integer
 , g_col_types       sys.odcinumberlist
 , g_col_delim       varchar2(1)
 , g_row_delim       varchar2(1)
 , g_pos1            integer
 , g_pos2            integer
 , g_col_num         integer
 , g_row_num         integer
 
 , static function parser(
      p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return anydataset pipelined
     using delimited_col_row
 
 , static function odcitabledescribe(
      p_tabtype   out   anytype
    , p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return number
 
 , static function odcitableprepare(
      p_context   out   delimited_col_row
    , p_tabinfo   in    sys.odcitabfuncinfo
    , p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return number
 
 , static function odcitablestart(
      p_context   in out delimited_col_row
    , p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return number
 
 , member function odcitablefetch(
      self      in out delimited_col_row
    , p_numrows   in    number
    , p_tab       out   anydataset
   ) return number
 
 , member function odcitableclose(
      self      in    delimited_col_row
   ) return number
)
/

create or replace type body delimited_col_row as
 
   static function odcitabledescribe(
      p_tabtype   out   anytype
    , p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return number
   is
      l_type      anytype;
      l_numcols   integer;
       
      l_pos1      pls_integer;
      l_pos2      pls_integer;
      l_pos3      pls_integer;
      l_colname   varchar2(1932);
      l_coltype   varchar2(1932);
   begin
      anytype.begincreate(dbms_types.typecode_object, l_type);
 
      l_pos1 := 0;
      loop
         l_pos2 := instr(p_cols, p_col_delim, l_pos1+1);
         l_pos3 := nvl(nullif(instr(p_cols, p_row_delim, l_pos1+1),0),length(p_cols)+1);
 
         l_colname := substr(p_cols, l_pos1+1, l_pos2-l_pos1-1);
         l_coltype := upper(substr(p_cols, l_pos2+1, l_pos3-l_pos2-1));
 
         if l_colname like '"%"' then
            l_colname := trim(both '"' from l_colname);
         else
            l_colname := upper(l_colname);
         end if;
          
         case
            when l_coltype like 'VARCHAR2(%)' then
               l_type.addattr(
                  l_colname
                , dbms_types.typecode_varchar2
                , null
                , null
                , to_number(trim(substr(l_coltype,10,length(l_coltype)-10)))
                , null
                , null
               );
 
            when l_coltype = 'NUMBER' then
               l_type.addattr(
                  l_colname
                , dbms_types.typecode_number
                , null
                , null
                , null
                , null
                , null
               );
 
            when l_coltype = 'DATE' then
               l_type.addattr(
                  l_colname
                , dbms_types.typecode_date
                , null
                , null
                , null
                , null
                , null
               );
 
            else
               raise dbms_types.invalid_parameters;
         end case;
          
         exit when l_pos3 > length(p_cols);
         l_pos1 := l_pos3;
      end loop;
       
      l_type.endcreate;
 
      anytype.begincreate(dbms_types.typecode_table, p_tabtype);
 
      p_tabtype.setinfo(
         null
       , null
       , null
       , null
       , null
       , l_type
       , dbms_types.typecode_object
       , 0
      );
 
      p_tabtype.endcreate();
 
      return odciconst.success;
 
   exception
      when others then
         return odciconst.error;
   end odcitabledescribe;

   static function odcitableprepare(
      p_context   out   delimited_col_row
    , p_tabinfo   in    sys.odcitabfuncinfo
    , p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return number
   is
      l_type      anytype;
      l_numcols   integer;
      l_col_types sys.odcinumberlist := sys.odcinumberlist();
    
      l_typecode  pls_integer;
      l_prec      pls_integer;
      l_scale     pls_integer;
      l_len       pls_integer;
      l_csid      pls_integer;
      l_csfrm     pls_integer;
      l_aname     varchar2(30);
      l_schemaname varchar2(30);
      l_typename  varchar2(30);
      l_version   varchar2(30);
      l_dummytype anytype;
   begin
      l_typecode := p_tabinfo.rettype.getattreleminfo(
                 1
               , l_prec
               , l_scale
               , l_len
               , l_csid
               , l_csfrm
               , l_type
               , l_aname
              );
    
      l_typecode := l_type.getinfo(
                 l_prec
               , l_scale
               , l_len
               , l_csid
               , l_csfrm
               , l_schemaname
               , l_typename
               , l_version
               , l_numcols
              );
       
      l_col_types.extend(l_numcols);
        
      for idx in 1..l_numcols loop
         l_col_types(idx) := l_type.getattreleminfo(
                    idx
                  , l_prec
                  , l_scale
                  , l_len
                  , l_csid
                  , l_csfrm
                  , l_dummytype
                  , l_aname
                 );
      end loop;
       
      p_context := delimited_col_row(
                      p_text
                    , l_type
                    , l_numcols
                    , l_col_types
                    , p_col_delim
                    , p_row_delim
                    , 0
                    , 0
                    , 0
                    , 0
                   );
    
      return odciconst.success;
   end odcitableprepare;

   static function odcitablestart(
      p_context   in out delimited_col_row
    , p_text      in    varchar2
    , p_cols      in    varchar2
    , p_col_delim in    varchar2 default '|'
    , p_row_delim in    varchar2 default ';'
   ) return number
   is
   begin
      p_context.g_buffer := p_text;
       
      p_context.g_pos1    := 0;
      p_context.g_pos2    := 0;
      p_context.g_col_num := 0;
      p_context.g_row_num := 0;
    
      return odciconst.success;
   end odcitablestart;

   member function odcitablefetch(
      self        in out delimited_col_row
    , p_numrows   in    number
    , p_tab       out   anydataset
   ) return number
   is
      l_row_cnt   integer := 0;
   begin
      if self.g_pos2 < length(self.g_buffer) then
         anydataset.begincreate(dbms_types.typecode_object, self.g_type, p_tab);
          
         loop
            self.g_col_num := self.g_col_num + 1;
             
            if self.g_col_num = 1 then
               l_row_cnt      := l_row_cnt + 1;
               self.g_row_num := self.g_row_num + 1;
               p_tab.addinstance;
               p_tab.piecewise();
            end if;
    
            if self.g_col_num < self.g_numcols then
               self.g_pos2 := instr(self.g_buffer, self.g_col_delim, self.g_pos1+1);
            else
               self.g_pos2 := nvl(nullif(instr(self.g_buffer, self.g_row_delim, self.g_pos1+1),0),length(self.g_buffer)+1);
            end if;
             
            case self.g_col_types(self.g_col_num)
               when dbms_types.typecode_varchar2 then
                  p_tab.setvarchar2(substr(self.g_buffer, self.g_pos1+1, self.g_pos2-self.g_pos1-1));
                
               when dbms_types.typecode_number then
                  p_tab.setnumber(to_number(substr(self.g_buffer, self.g_pos1+1, self.g_pos2-self.g_pos1-1)));
                
               when dbms_types.typecode_date then
                  p_tab.setdate(to_date(substr(self.g_buffer, self.g_pos1+1, self.g_pos2-self.g_pos1-1)));
            end case;
             
            exit when self.g_pos2 > length(self.g_buffer);
            self.g_pos1 := self.g_pos2;
    
            if self.g_col_num = self.g_numcols then
               self.g_col_num := 0;
               exit when l_row_cnt >= p_numrows;
            end if;
         end loop;
    
         p_tab.endcreate;
      end if;
      return odciconst.success;
   end odcitablefetch;

   member function odcitableclose(
      self      in    delimited_col_row
   ) return number
   is
   begin
      return odciconst.success;
   end odcitableclose;
 
end;
/

/*
   Using ODCI type to parse the delimited text
*/
select category, beverage_type, color
  from beverages2
     , table(
          delimited_col_row.parser(
             typelist
           , 'BEVERAGE_TYPE|VARCHAR2(10);COLOR|VARCHAR2(10)'
           , '|'
           , ';'
          )
       ) type_and_color
 order by category, beverage_type;

/*
   Cleanup
*/
drop type delimited_col_row;
drop table beverages2 purge;


/* ------------------------  \
|  ROLL rows to string       |
\  ------------------------ */

/*
   Table with test data
*/
create table beverages3 (
   category       varchar2(10)
 , beverage_type  varchar2(10)
);

insert into beverages3 values ('Beer', 'Pilsner');
insert into beverages3 values ('Beer', 'Ale');
insert into beverages3 values ('Beer', 'Stout');
insert into beverages3 values ('Wine', 'Red');
insert into beverages3 values ('Wine', 'Champagne');

commit;

select *
  from beverages3
 order by category, beverage_type;

/*
   LISTAGG
*/
select category
     , listagg(beverage_type, ';') within group (
          order by beverage_type
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   LISTAGG ON OVERFLOW (v.12.2)
*/
column category format a8
column typelist format a29

select category
     , listagg(
          beverage_type, ';'
          on overflow truncate with count
       ) within group (
          order by lvl, beverage_type
       ) typelist
  from beverages3
  cross join (
      select level lvl from dual
      connect by level <= 1000
  )
 group by category
 order by category;

/*
   LISTAGG support for DISTINCT (v.19)
*/

select category
     , listagg(
          distinct
          beverage_type, ';'
       ) within group (
          order by beverage_type
       ) typelist
  from beverages3
  cross join (
      select level lvl from dual
      connect by level <= 1000
  )
 group by category
 order by category;

/*
   COLLECT
*/
create type beverage_collection_type as table of varchar2(10);
/

create or replace function beverage_typecoll_to_string (
   typecoll  in  beverage_collection_type
)
   return varchar2
is
   type_string varchar2(4000);
begin
   for idx in typecoll.first .. typecoll.last loop
      if idx = typecoll.first then
         type_string := typecoll(idx);
      else
         type_string := type_string || ';' || typecoll(idx);
      end if;
   end loop;
   return type_string;
end beverage_typecoll_to_string;
/

select category
     , beverage_typecoll_to_string(
          cast(
             collect(
                beverage_type
                order by beverage_type
             )
             as beverage_collection_type
          )
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   APEX_STRING.JOIN
*/

select category
     , apex_string.join(
          cast(
             collect(
                beverage_type
                order by beverage_type
             )
             as apex_t_varchar2
          )
        , ';'
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   CLOB support
*/

select category
     , apex_string.join_clob(
          cast(
             collect(
                beverage_type
                order by beverage_type
             )
             as apex_t_varchar2
          )
        , ';'
        , 12 /* dbms_lob.call */
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   ODCI implementation of string aggregate function by Tom Kyte
   https://asktom.oracle.com/pls/apex/f?p=100:11:0::::P11_QUESTION_ID:2196162600402
*/

create or replace type string_agg_type as object
(
   total varchar2(4000),

   static function
        ODCIAggregateInitialize(sctx IN OUT string_agg_type )
        return number,

   member function
        ODCIAggregateIterate(self IN OUT string_agg_type ,
                             value IN varchar2 )
        return number,

   member function
        ODCIAggregateTerminate(self IN string_agg_type,
                               returnValue OUT  varchar2,
                               flags IN number)
        return number,

   member function
        ODCIAggregateMerge(self IN OUT string_agg_type,
                           ctx2 IN string_agg_type)
        return number
);
/

create or replace type body string_agg_type
is

   static function ODCIAggregateInitialize(sctx IN OUT string_agg_type)
   return number
   is
   begin
       sctx := string_agg_type( null );
       return ODCIConst.Success;
   end;

   member function ODCIAggregateIterate(self IN OUT string_agg_type,
                                        value IN varchar2 )
   return number
   is
   begin
       self.total := self.total || ';' || value;
       return ODCIConst.Success;
   end;

   member function ODCIAggregateTerminate(self IN string_agg_type,
                                          returnValue OUT varchar2,
                                          flags IN number)
   return number
   is
   begin
       returnValue := ltrim(self.total,';');
       return ODCIConst.Success;
   end;

   member function ODCIAggregateMerge(self IN OUT string_agg_type,
                                      ctx2 IN string_agg_type)
   return number
   is
   begin
       self.total := self.total || ctx2.total;
       return ODCIConst.Success;
   end;

end;
/

create or replace function stragg(input varchar2 )
   return varchar2
   parallel_enable aggregate using string_agg_type;
/

/*
   Using ODCI stragg() - note: no order by
*/
select category
     , stragg(beverage_type) typelist
  from beverages3
 group by category
 order by category;

/*
   XMLAGG
*/

select category
     , rtrim(
          xmlagg(
             xmlelement(z, beverage_type, ';')
             order by beverage_type
          ).extract('//text()').getstringval()
        , ';'
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   Skip extraction by claiming content is xml
*/

select category
     , rtrim(
          xmlagg(
             xmlparse(
                content beverage_type || ';' wellformed
             )
             order by beverage_type
          ).getstringval()
        , ';'
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   Supports CLOB in case of large results
*/

select category
     , rtrim(
          xmlagg(
             xmlparse(
                content beverage_type || ';' wellformed
             )
             order by beverage_type
          ).getclobval()
        , ';'
       ) typelist
  from beverages3
 group by category
 order by category;

/*
   Cleanup
*/
drop function stragg;
drop type string_agg_type;
drop function beverage_typecoll_to_string;
drop type beverage_collection_type;
drop table beverages3 purge;


/* =========================================  \
|  End of script                              |
\  ========================================= */
